#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 6/29/19
"""
.. currentmodule:: bnrml.normalizers
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Be normal!
"""
from abc import ABC, abstractmethod
from functools import lru_cache
from collections import OrderedDict
from enum import Enum
import hashlib
from pathlib import Path
import string
from typing import (
    Any,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    NamedTuple,
    Union
)
from frozenordereddict import FrozenOrderedDict
from pytypewriter import Exportable, pycls, pyfqn
from . import mappings
from .errors import BnrmlException


_ZERO_PUNCTUATION = str.maketrans(
    '', '',
    string.punctuation
)  #: string transform that eliminates punctuation


class Normal(Exportable):
    """
    This is a data object that represents a "normal" value along with all of
    its abbreviations and aliases.
    """
    __slots__ = ['_normal', '_abbreviations', '_aliases', 'hash_']

    class Components(Enum):
        """
        This is an enumeration of the principal components of a `Normal`.
        """
        NORMAL = 'NORMAL'  #: the normal form
        ABBREVIATION = 'ABBREVIATION'  #: an abbreviation
        ALIAS = 'ALIAS'  #: an alias

    def __init__(
            self,
            normal: str,
            abbreviation: str,
            aliases: Iterable[str]
    ):
        """

        :param normal: the normal form
        :param abbreviation: the abbreviated form
        :param aliases: aliases
        """
        self._normal = normal
        self._abbreviation = abbreviation
        self._aliases = frozenset(aliases)
        self.hash_: int or None = None

    @property
    def normal(self) -> Union[str, int]:
        """
        Get the normal form.
        """
        return self._normal

    @property
    def abbreviation(self) -> str:
        """
        Get the abbreviated form
        """
        return self._abbreviation

    @property
    def aliases(self) -> FrozenSet[str]:
        """
        Get the aliases.
        """
        return self._aliases

    def export(self) -> Mapping[str, Any]:
        """
        Export the instance as a mapping of simple types.

        :return: the mapping
        """
        return {
            'normal': self._normal,
            'abbreviation': self._abbreviation,
            'aliases': list(self._aliases)
        }

    @classmethod
    def load(cls, data: Mapping[str, Any]) -> 'Normal':
        """
        Create an instance from a mapping.

        :param data: the data
        :return: the instance
        """
        # Construct the instance.
        obj = cls(
            **{
                k: data.get(k) for k in [
                    'normal',
                    'abbreviation',
                    'aliases'
                ]
            }
        )
        # If the data contains a ready-made hash, let's apply it now.
        hash_ = data.get('hash')
        if hash_ is not None:
            setattr(obj, 'hash_', hash_)
        # That's that.
        return obj

    @classmethod
    def loadd(cls, normal: str, data: Mapping[str, Any]) -> 'Normal':
        """
        Create an instance from normal value and a mapping of other
        attribute values (as you might find them in a dictionary that utilizes
        the normal value as a key and the other attributes as a value object).

        :param normal: the normal form (key)
        :param data: the object defining other attributes
        :return:
        """
        return cls.load(
            {
                'normal': normal,
                **data
            }
        )

    def _makehash(self) -> int:
        """
        Generate a hash for the datum.

        :return: the hash
        """
        md5 = hashlib.sha3_512(  # pylint: disable=no-member
            str(
                {
                    'normal': self._normal,
                    'abbreviation': self._abbreviation,
                    'aliases': list(self._aliases),
                }
            ).encode('UTF-8')
        )
        return int(md5.hexdigest(), 16)

    def __hash__(self):
        if self.hash_ is None:
            self.hash_ = self._makehash()
        return self.hash_

    def __eq__(self, other):
        return self.hash_ == getattr(other, 'hash_', None)

    def __ne__(self, other):
        return not self.hash_


class IndexKeyViolation(NamedTuple):
    """
    This is a data object that describes a key conflict (duplicate values)
    in a set of normal definitions.

    .. seealso::

        :py:class:`IndexKeyViolationException`
    """
    key: str
    component: Normal.Components


class IndexKeyViolationException(BnrmlException):
    """
    This exception is raised in response to attempts to index multiple normal
    definitions that have conflicting keys (duplicate values).
    """
    def __init__(
            self,
            message: str,
            violations: Iterable[IndexKeyViolation],
            inner: Exception = None
    ):
        """

        :param message: the exception message
        :param violations: the individual violations
        :param inner: the exception that caused this exception
        """
        super().__init__(message=message, inner=inner)
        self._violations = list(violations) if violations else []

    @property
    def violations(self) -> Iterable[IndexKeyViolation]:
        """
        Get the individual violations.
        """
        return iter(self._violations)


class NormalsIndex:
    """
    This is an indexer for :py:class:`Normal` definitions.
    """
    def __init__(self, normals: Iterable[Normal] = None):
        """

        :param normals: the initial normals contained in the index
        """
        #: the full set of normals
        self._normals = set(normals) if normals else set()
        #: normals indexed by the normal form
        self._by_normal: Dict[str, Normal] = {}
        #: normals indexed by abbreviations
        self._by_abbrev: Dict[str, Normal] = {}
        #: normals indexed by alias
        self._by_alias: Dict[str, Normal] = {}
        # If normal definitions were supplied...
        if self._normals:
            # ...update now.
            self.update()

    @property
    def normals(self) -> Iterable[Normal]:
        """
        Get the normals present in the index.
        """
        return iter(self._normals)

    @classmethod
    def std_key(cls, key: str) -> str:
        """
        Standardize an index key value.

        :param key: the key
        :return: the standard form
        """
        # Remove punctuation.
        _key = key.lower().translate(_ZERO_PUNCTUATION).strip()
        # Compress whitespace.
        return ' '.join(_key.split())

    def update(self, normals: Iterable[Normal] = None) -> 'NormalsIndex':
        """
        Generate the indexes.

        :return: this instance
        """
        # Create new data structures and indexes.
        _normals = (
            (
                self._normals | set(normals) if normals else set()
            )
            if normals else
            self._normals
        )
        by_normal = {}
        by_abbrev = {}
        by_alias = {}
        all_keys: Dict[str, Normal] = {}

        # We'll want to keep track of index violations.
        violations: List[IndexKeyViolation] = []
        for normal in _normals:
            # Index the normal forms, abbreviations, and aliases.
            for index, component, keys in [
                    (
                        by_normal,
                        Normal.Components.NORMAL,
                        [normal.normal]
                    ),
                    (
                        by_abbrev,
                        Normal.Components.ABBREVIATION,
                        [normal.abbreviation]
                    ),
                    (
                        by_alias,
                        Normal.Components.ALIAS,
                        normal.aliases
                    )
            ]:
                for key in keys:
                    # If the key is empty...
                    if not key:
                        # ...move along.
                        continue
                    # Standardize the key.
                    _key = self.std_key(key)
                    # If we've already seen this key but it's associated
                    # with another `Normal`...
                    if all_keys.get(_key) != normal:
                        violations.append(
                            IndexKeyViolation(
                                key=_key,
                                component=component
                            )
                        )
                        continue
                    else:
                        # Make note that we've seen this key.
                        all_keys[key] = normal
                        # Simply add this one to the index.
                        index[_key] = normal
        # If any violations were detected...
        if violations:
            # ...raise an exception.
            raise IndexKeyViolationException(
                message=f'Index violations exist in the series: {violations}',
                violations=violations
            )

        # It looks like we're good, so replace the existing data structures and
        # indexes.
        self._normals = _normals
        self._by_normal = by_normal
        self._by_abbrev = by_abbrev
        self._by_alias = by_alias

        # Return this instance to the caller.
        return self

    def search(self, s: str) -> Normal or None:
        """
        Search for a normal definition.

        :param s: the the denormalized string
        :return: the normal definition or `None`
        """
        _std = self.std_key(s)
        for idx in [self._by_normal, self._by_abbrev, self._by_alias]:
            try:
                return idx[_std]
            except KeyError:
                pass
        return None


class Normalization(Exportable):
    """
    A `Normalization` contains the results of an attempt to normalize a value.
    """
    __slots__ = [
        '_denormalized',
        '_normalized',
        '_normals',
        '_extras'
    ]

    def __init__(
            self,
            denormalized: str,
            normalized: str,
            normals: Iterable[Normal] = None,
            extras: OrderedDict = None
    ):
        """

        :param denormalized: the denormalized value
        :param normalized: the normal form
        :param normals: the :py:class:`Normal` definitions that match the
            denormalized value (if any)
        :param extras: extra values
        """
        self._denormalized = denormalized
        self._normalized = normalized
        self._normals = list(normals) if normals else []
        self._extras = (
            FrozenOrderedDict(extras) if extras else FrozenOrderedDict()
        )

    @property
    def denormalized(self) -> str:
        """
        Get the denormalized value.
        """
        return self._denormalized

    @property
    def normalized(self) -> str:
        """
        Get the normal form.
        """
        return self._normalized

    @property
    def normals(self) -> Iterable[Normal]:
        """
        Get the :py:class:`Normal` definitions that matched the denormalized
            value (if there were any)
        """
        return iter(self._normals)

    @property
    def extras(self) -> FrozenOrderedDict:
        """
        Get the extra values.
        """
        return self._extras

    def export(self) -> Mapping[str, Any]:
        """
        Export the instance as a mapping of simple types.

        :return: the mapping
        """
        return {
            k: v for k, v in
            {
                **{slot[1:]: getattr(self, slot) for slot in self.__slots__},
                'normals': [n.export() for n in self._normals],
                'extras': dict(self.extras)
            }.items()
        }

    @classmethod
    def load(cls, data: Mapping[str, Any]) -> Any:
        """
        Create an instance from a mapping.

        :param data: the data
        :return: the instance
        """
        return cls(**data)


def load_normals(data: Mapping[str, Any]) -> Iterable[Normal]:
    """
    Parse :py:class:`Normals <Normal>` from a file.

    :param data: the data mapping that describes the `Normals`
    :return: the normals described in the file
    """
    # If there is some data to work with...
    if data:
        # ...go through all the keys in the dictionary.  (We'll consider those
        # keys to be the normal form.)
        for k, v in data.items():
            # Create the `Normal` from the data at hand...
            normal = Normal(
                normal=k,
                abbreviation=v.get('abbreviation'),
                aliases=v.get('aliases')
            )
            # ...and yield it to the caller.
            yield normal


def loadf_normals(path: Path or str) -> Iterable[Normal]:
    """
    Load :py:class:`Normals <Normal>` from a file.

    :param path: the file path
    :return: the normals described in the file
    """
    _path = (
        path if isinstance(path, Path) else Path(path)
    ).expanduser().resolve()
    data = mappings.loadf(path=_path)
    return load_normals(data=data.get('normals'))


class Normalizer(Exportable, ABC):
    """
    Normalizers convert values to their normal forms (if possible).
    """

    def __init__(
            self,
            **kwargs
    ):
        """

        :param normals: the set of normals
        :param kwargs: additional keyword arguments
        """
        self._kwargs: Dict[str, Any] = kwargs  #: the subclass keyword arguments

    @abstractmethod
    def normalize(
            self,
            s: str,
            default: Union[str, Normal, None] = None
    ) -> Normalization or None:
        """
        Normalize a value.

        :param s: the value to normalize
        :param default: the default value to return if no normal value applies
        :return: the normalized value
        """

    def export(self) -> Mapping[str, Any]:
        """
        Export the instance as a mapping of simple types.

        :return: the mapping
        """
        return {
            'normalizer': {
                'type': pyfqn(self),
                'args': {
                    k: v for k, v in self._kwargs
                }
            }
        }

    @classmethod
    def load(cls, data: Mapping[str, Any]) -> Any:
        """
        Create an instance from a mapping.

        :param data: the data
        :return: the instance
        """
        # Get the `normalizer` value from the data.
        normalizer_data = data['normalizer']
        # Let's figure out what type of `Normalizer` we're going to create.
        _type = normalizer_data.get('type')
        # Let's figure out what type we're supposed to load.
        _cls = pycls(_type) if _type else cls
        # If its not this class...
        if _cls is not cls:
            # ...let the other class' load method take it from here.
            return _cls.load(data)
        # Otherwise, parse out the arguments...
        _args = normalizer_data.get('args', ())
        # ...and create an instance.
        return _cls(**_args)

    @classmethod
    def loadf(cls, path: Union[Path, str]) -> 'Normalizer':
        """
        Load a normalizer from a configuration file.

        :param path: the path to the configuration file
        :return: the normalizer
        """
        data = mappings.loadf(path)
        return cls.load(data)


class SimpleNormalizer(Normalizer, ABC):
    """
    This is an implementation of :py:class:`Normalizer` that provides standard
    logic for working with a pre-defined set of :py:class:`Normals <Normal>`
    in a single index.
    """

    def __init__(
            self,
            normals: Union[Iterable[Normal], Path, str] = None,
            prefer: Normal.Components = Normal.Components.NORMAL,
            index_type: Union[type, str] = NormalsIndex,
            **kwargs
    ):
        """

        :param normals: the set of normals or the path to a file that defines
            the normals
        :param prefer: indicates a preference for using the
            :py:attr:`normal <bnrml.normalizers.Normal.Components.NORMAL>`
            form (the default) or the
            :py:attr:`abbreviation <bnrml.normalizers.Normal.Components.ABBREVIATION>`
            as `normalized`
        :param index_type: the :py:class:`NormalsIndex <NormalsIndex>` type
        :param kwargs: additional keyword arguments
        """
        # Pass the keyword arguments to the parent.
        super().__init__(**kwargs)
        # Get the normalization preference.
        self._prefer: Normal.Components = prefer
        # If the preference isn't either `NORMAL` or `ABBREVIATION`, it's not
        # a valid value.
        if self._prefer not in [
                Normal.Components.NORMAL,
                Normal.Components.ABBREVIATION
            ]:
            raise ValueError(
                f"The `prefer` parameter must be {Normal.Components.NORMAL} or "
                f"{Normal.Components.ABBREVIATION}"
            )
        # Extract the normals (if any) from the constructor args.
        self._normals: List[Normal] = (
            list(loadf_normals(path=normals))
            if isinstance(normals, (Path, str))
            else list(normals)
        ) if normals else []
        # Let's figure out what we're using for a `NormalsIndex`.
        _index_type = (
            index_type if isinstance(index_type, type)
            else pycls(index_type)
        )
        # Generate the index from the normals that were provided.
        self._nindex: NormalsIndex = _index_type().update(self._normals)

    @lru_cache()
    def normalize(
            self,
            s: str,
            default: Union[str, Normal, None] = None
    ) -> Normalization or None:
        """
        Normalize a value.

        :param s: the value to normalize
        :param default: the default value to return if no normal value applies
        :return: the normalized value
        """
        # Let the index do it's work.
        normal = self._nindex.search(s)
        # If the index returned a `Normal`...
        if normal:
            # The preferred value to return as the "normalized" value may be the
            # `normal`, but if the normalizer has been configured to prefer the
            # abbreviation, we'll use that.
            _normalized = (
                normal.abbreviation
                if self._prefer == Normal.Components.ABBREVIATION
                else normal.normal
            )
            # Just in case this normalizer prefers the abbreviation, but the
            # selected normal doesn't actually have an abbreviation (resulting
            # in an empty value for `_normalized`)...
            if not _normalized:
                # Let's resort to the safe normal form.
                _normalized = normal.normal
            # Now we can construct a normalization.
            return Normalization(
                denormalized=s,
                normalized=_normalized,
                normals=[normal]
            )
        # We need to resort to the default, which may be `None`...
        if default is None:
            return None
        # If the `default` is a string...
        if isinstance(default, str):
            # ...we can construct a `Normalization` with no `Normal` instances.
            return Normalization(denormalized=s, normalized=s)
        # If the `default` is a `Normal`...
        if isinstance(default, Normal):
            # ...we can construct a `Normalization` using that information.
            return Normalization(
                denormalized=s,
                normalized=default.normal,
                normals=[default]
            )
        # If we get to this point it means that the `default` parameter isn't
        # of a type we can use here.
        raise TypeError(
            f"The type of the 'default' parameter ({type(default).__name__}) "
            f"is invalid."
        )

    def export(self) -> Mapping[str, Any]:
        """
        Export the instance as a mapping of simple types.

        :return: the mapping
        """
        data = dict(super().export())
        data['normalizer']['args']['normals'] = {
            n.normal: {
                k: v for k, v
                in n.export().items()
                # The normal form is the key.  Don't repeat it.
                if k != 'normal'
            }
            for n in self._normals
        }
        return data

    @classmethod
    def _load_parse_args(cls, args: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        Subclasses may override this method to parse arguments constructor
        arguments as they are parsed in the :py:funct:`load` method.

        :param args: the argument dictionary
        :return: the modified argument dictionary
        """
        return args

    @classmethod
    def load(cls, data: Mapping[str, Any]) -> Any:
        """
        Create an instance from a mapping.

        :param data: the data
        :return: the instance
        """
        # Get the 'normalizer' key wherein the configuration resides.
        normalizer_data = data['normalizer']
        # Let's see what type we're working with.
        _type = normalizer_data.get('type')
        # If no type is specified, assume the current type.
        _cls = pycls(_type) if _type else cls
        # Get the arguments.
        _args = normalizer_data.get('args', ())
        # Let's see if there is a 'normals' argument.  If so, it'll be
        # complex...
        _normals = _args.get('normals', {})
        # ...so we need to do some additional work to parse them into
        # `Normal` instances.
        _args['normals'] = [
            Normal.loadd(normal, data)
            for normal, data
            in _normals.items()
        ]
        # Let's see if the args contain the `prefer` argument.
        _prefer = _args.get('prefer')
        if _prefer:
            try:
                _args['prefer'] = Normal.Components[str(_prefer).upper()]
            except KeyError:
                raise ValueError(
                    f'{repr(str(_prefer))} is not a valid argument for the '
                    f"'prefer' parameter."
                )
        # Give subclasses a change to further process the arguments.
        _args = cls._load_parse_args(_args)
        # That's that.
        return _cls(**_args)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 7/13/19
"""
.. currentmodule:: bnrml.mappings
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains utility functions that may be helpful when working with
mappings (y'know... like dictionaries).
"""
import json
from pathlib import Path
from typing import Mapping
import toml

#: supported file extensions for mapping files
SUPPORTED_MAPPING_FILE_EXTS = ('.toml', '.zip')


def loads(s: str) -> Mapping:
    """
    Convert a `JSON` or `TOML` string into a `Mapping`

    :param s: the string
    :return: the data structure represented by the string
    :raises ValueError: if the string isn't a proper example of a recognized
        format
    """
    # Try to load the data as JSON.
    try:
        return json.loads(s)
    except json.JSONDecodeError:
        # It doesn't look like it's JSON encoded.
        pass
    # Try to load the data as TOML.
    try:
        return toml.loads(s)
    except toml.TomlDecodeError:
        # It doesn't seem to be TOML either.
        raise ValueError(
            'The format is not supported.  Supported formats are JSON '
            'and TOML.'
        )


def loadf(path: Path or str) -> Mapping:
    """
    Convert a `JSON` or `TOML` file to a mapping

    :return: the data structure represented by the string
    :raises ValueError: if the string isn't a proper example of a recognized
        format
    """
    _path = (
        path if isinstance(path, Path)
        else Path(path)
    ).expanduser().resolve()
    return loads(_path.read_text())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 9/9/19
"""
.. currentmodule:: bnrml.soundalike.normalizers
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This module contains extensions for normalizers the perform fuzzy matching.
"""
from typing import Dict, Iterable, List, NamedTuple, Union
from . import distance, encoding
from .distance import TextDistance
from .encoding import FuzzyEncoder
from ..normalizers import Normal, NormalsIndex


class SourceNormal(NamedTuple):
    """Associates a ``Normal`` definition with a source string."""
    source: str  #: a source string
    normal: Normal  #: a normal associated with the source string


class DistanceNormal(NamedTuple):
    """Describes the distance between a ``Normal`` and its source string."""
    #: the calculated distance from a string to one of the keys in a normal
    #: definition
    distance: float
    source_normal: SourceNormal


class SoundalikeNormalsIndex(NormalsIndex):
    """
    This is a `Normal` indexer that attempts fuzzy matching in cases where no
    direct match is found.
    """
    def __init__(
            self,
            normals: Iterable[Normal] = None,
            encoder: Union[FuzzyEncoder, str] = 'soundex',
            text_distance: Union[distance.TextDistance, str] = 'levenshtein'
    ):
        # Let the superclass do its thing.
        super().__init__(normals=normals)
        # Figure out what we'll use to encode strings.
        self._encoder = (
            encoder if isinstance(encoder, FuzzyEncoder)
            else encoding.encoder(encoder)()
        )  #: the principal encoder
        self._text_distance = (
            text_distance if isinstance(text_distance, TextDistance)
            else distance.text_distance(text_distance)()
        )  #: the text distance utility
        #: an index of soundalike-encoded strings (keys) mapped to the
        #: the source strings and normal definitons from which the strings
        #: came
        self._by_code: Dict[str, List[SourceNormal]] = {}

    def update(self, normals: Iterable[Normal] = None) -> 'NormalsIndex':
        """
        Generate the indexes.

        :return: this instance
        """
        # Let the superclass perform its work.
        super().update(normals)
        # Create an empty index.
        by_code: Dict[str, List[SourceNormal]] = {}

        # Let's go through all the normals.
        for normal in self.normals:
            # Collect all the unique values.
            sources = set([normal.normal] + [a for a in normal.aliases])
            # Now let's go through all the sources (which include the normal
            # form and all the aliases).
            for source in sources:
                # If there is, somehow, an empty value...
                if not source:
                    continue  # ...just move along.
                # Lowercase the source string and encode it.
                code = self._encoder.encode(self.std_key(source))
                # If no usable code was generated...
                if not code:
                    continue  # ...skip it.
                # Create a `SourceNormal` data object to keep track of the
                # fact that this source is associated with this normal
                # definition.
                source_normal = SourceNormal(
                    source=source,
                    normal=normal
                )
                # We may have seen this code before, so let's try to just
                # append this value...
                try:
                    by_code[code].append(source_normal)
                except KeyError:
                    # ...but if we haven't, we need to start a new list...
                    source_normals = [source_normal]
                    # ...and file it under this code as the key.
                    by_code[code] = source_normals

        # When we're finished, we can replace the existing index.
        self._by_code = by_code
        # The caller expects us to return this instance.
        return self

    def search(self, s: str) -> Normal or None:
        """
        Search for a normal definition.

        :param s: the denormalized string
        :return: the normal definition or `None`
        """
        # Let the superclass take a shot.
        normal = super().search(s)
        # If the superclass found something...
        if normal:
            # ...that's our answer.
            return normal

        # Standardize the input.
        s_std = self.std_key(s)

        # Encode the string to its fuzzy matching hash.
        code = self._encoder.encode(s_std)

        # Get all the normal definitions that are indexed under this
        # fuzzy matching hash code.
        soundalikes: Iterable[SourceNormal] = self._by_code.get(code)

        # If we don't recognize the code at all...
        if not soundalikes:
            # ...we don't have a match.
            return None

        # Sort the soundalikes according to their distance from the original
        # string.
        soundalikes_sorted: List[DistanceNormal] = sorted(
            [
                DistanceNormal(
                    distance=self._text_distance.distance(s_std, sn.source),
                    source_normal=sn
                )
                for sn in soundalikes
            ],
            key=lambda dn: dn.distance
        )
        # Return the normal definition that sorted to the top.
        return soundalikes_sorted[0].source_normal.normal



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 9/8/19
"""
.. currentmodule:: bnrml.soundalike.distance
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Just how similar are two strings?
"""
from abc import ABC
from enum import Enum
import string
from typing import Callable, Type
import textdistance
from pytypewriter import pycls, pyfqn, InvalidTypeException
from ..errors import BnrmlException


_ZERO_PUNCTUATION = str.maketrans(
    '', '',
    string.punctuation
)  #: string transform that eliminates punctuation


class UnknownTextDistanceException(BnrmlException):
    """
    Raised when a caller attempts to retrieve an unknown
    :py:class:`TextDistance <TextDistance>`

    .. seealso::

        :py:func:`text_distance`
    """


class TextDistances(Enum):
    """
    This is an enumeration of the known, built-in text measuring rods.
    """
    HAMMING = 'HAMMING'
    LEVENSHTEIN = 'LEVENSHTEIN'


class TextDistance(ABC):
    """
    This is a utility object that measures the "distance" (or difference)
    between two words.
    """

    #: indicates a text distance beyond the threshold
    OUT_OF_RANGE: float = float('inf')

    def __init__(
            self,
            distance_fn: Callable[[str, str], int],
            ndistance_fn: Callable[[str, str], float],
            threshold: int
    ):
        """

        :param distance_fn: a function that measures the quantitative text
            distance between two words
        :param ndistance_fn: a function that measures the distance between two
            words and returns a value between 0 and 1 where 0 means the words
            are the same and 1 means they are completely different
        :param threshold: the maximum quantitative distance between two words
            before they are considered completely distinct
        """
        self._distance_fn: Callable[[str, str], int] = distance_fn
        self._ndistance_fn: Callable[[str, str], float] = ndistance_fn
        self._threshold: int = threshold

    @property
    def threshold(self) -> int:
        """
        Get the maximum quantitative distance between two words before they are
        considered completely distinct.
        """
        return self._threshold

    @threshold.setter
    def threshold(self, value: int):
        """
        Set the maximum quantitative distance between two words before they are
        considered completely distinct.

        :param value: the threshold value
        """
        if value < 1:
            raise ValueError("'value' must be > 1.")
        if value >= self.OUT_OF_RANGE:
            raise ValueError(
                "'value' must not exceed the maximum (out of range) value."
            )
        self._threshold = value

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of ruler.
        """
        return pyfqn(cls)

    def distance(
            self,
            texta: str,
            textb: str,
            normalized: bool = False
    ) -> float:
        """
        Encode a string to its fuzzy matching hash.

        :param texta: a string
        :param textb: another string
        :param normalized: `True` to return a normalized value between 0 and
            1 where 0 means the same and 1 means completely different
        :return: the distance

        .. note::

            If the distance exceeds the :py:func:`threshold` value, a very large
            value equal to :py:attr:`OUT_OF_RANGE` is returned to the caller.
        """
        # Clean up the input strings.
        _a = texta.strip().lower().translate(_ZERO_PUNCTUATION)
        _b = textb.strip().lower().translate(_ZERO_PUNCTUATION)
        # Use the function supplied to get the distance.
        if normalized:
            return self._ndistance_fn(texta, textb)
        # Get the distance.
        d = float(self._distance_fn(texta, textb))
        # If the distance is within the threshold, return it.  Otherwise return
        # the "out-of-range" value.
        return d if d <= self._threshold else self.OUT_OF_RANGE


class Hamming(TextDistance):
    """
    This is a rule that returns the
    `Hamming distance <https://en.wikipedia.org/wiki/Hamming_distance>`_
    between two strings.
    """
    def __init__(self, threshold: int = 4):
        """

        :param threshold: the maximum quantitative distance between two words
            before they are considered completely distinct
        """
        super().__init__(
            distance_fn=textdistance.hamming.distance,
            ndistance_fn=textdistance.hamming.distance,
            threshold=threshold
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of ruler.
        """
        return TextDistances.HAMMING.name


class Levenshtein(TextDistance):
    """
    This is a rule that returns the
    `Levenshtein distance <hhttps://en.wikipedia.org/wiki/Levenshtein_distance>`_
    between two strings.
    """
    def __init__(self, threshold: int = 4):
        """

        :param threshold: the maximum quantitative distance between two words
            before they are considered completely distinct
        """
        super().__init__(
            distance_fn=textdistance.levenshtein.distance,
            ndistance_fn=textdistance.levenshtein.distance,
            threshold=threshold
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of ruler.
        """
        return TextDistances.LEVENSHTEIN.name


_TEXT_DISTANCE_TYPES = {
    **{
        k: v for k, v in {
            TextDistances.HAMMING: Hamming,
            TextDistances.LEVENSHTEIN: Levenshtein
        }.items()
    },
    **{
        t.key(): t for t in [
            Hamming,
            Levenshtein
        ]
    }
}  #: a mapping of encoder type keys to encoder types


def text_distance(key: TextDistances or str) -> Type:
    """
    Get a :py:class:`Rod` type that corresponds to a given enumeration
    value or name.

    :param key: the name
    :return: the corresponding type
    :raises KeyError: if no mapping exists for the key
    """
    try:
        return _TEXT_DISTANCE_TYPES[
            key.upper() if isinstance(key, str)
            else key
        ]
    except KeyError:
        try:
            _cls = pycls(key)
            return _cls()
        except InvalidTypeException:
            raise UnknownTextDistanceException(
                f"No class is associated with text distance key "
                f"'{key}'."
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 9/8/19
"""
.. currentmodule:: bnrml.soundalike.encoding
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Generate soundalike encodings for strings.
"""
from abc import ABC
from enum import Enum
from functools import partial
import re
import string
from typing import Callable, cast, Type
import phonetics
from pytypewriter import pyfqn


_ZERO_PUNCTUATION = str.maketrans(
    '', '',
    string.punctuation
)  #: string transform that eliminates punctuation

_WORD_DIV = re.compile(r'\s+')  #: identifies boundaries between words


class Encoders(Enum):
    """
    This is an enumeration of the known, built-in soundalike encoders.
    """
    SOUNDEX = 'SOUNDEX'
    NYSIIS = 'NYSIIS'
    METAPHONE = 'METAPHONE'
    DMETAPHONE = 'DMETAPHONE'


class FuzzyEncoder(ABC):
    """
    Use a "fuzzy" encoder to generate soundalike codes for strings.
    """
    def __init__(self, fn: Callable[[str], str]):
        """

        :param fn: the function that performs the soundalike
        """
        self._fn: Callable[[str], str] = fn

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of encoder.
        """
        return pyfqn(cls)

    def encode(self, s: str) -> str or None:
        """
        Encode a string to its fuzzy matching hash.

        :param s: the string
        :return: the encoding (or `None` if no encoding is possible)
        """
        # No input means no output.
        if not s:
            return None
        # Clean up the input string.
        _s = s.strip().lower().translate(_ZERO_PUNCTUATION)
        # Check again... if we were given an empty string, or just punctuation
        # we aren't going to generate a code.
        if not _s:
            return None
        # There may be multiple individual strings, so we may need to encode
        # multiple times.
        codes = [
            self._fn(i) for i in _WORD_DIV.split(_s)
        ]
        # Put it all back together.
        code = ' '.join(codes)
        # Return what we got.
        return code.upper() if code else None


class SoundexEncoder(FuzzyEncoder):
    """
    This is a fuzzy encoder that uses the
    `soundex <https://en.wikipedia.org/wiki/Soundex>`_ algorithm.
    """
    def __init__(self, size: int = 4):
        super().__init__(
            fn=cast(
                Callable[[str], str],
                partial(phonetics.soundex, size=size)
            )
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of encoder.
        """
        return Encoders.SOUNDEX.name


class NysiisEncoder(FuzzyEncoder):
    """
    This is a fuzzy encoder that uses the
    `NYSIIS <https://www.health.ny.gov/prevention/immunization/information_system/>`_
    algorithm.
    """
    def __init__(self):
        super().__init__(
            fn=cast(
                Callable[[str], str],
                phonetics.nysiis
            )
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of encoder.
        """
        return Encoders.NYSIIS.name


class MetaphoneEncoder(FuzzyEncoder):
    """
    This is a fuzzy encoder that uses the
    `metaphone <https://en.wikipedia.org/wiki/Metaphone>`_ algorithm.
    """
    def __init__(self):
        super().__init__(
            fn=cast(
                Callable[[str], str],
                phonetics.metaphone
            )
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of encoder.
        """
        return Encoders.METAPHONE.name


class DMetaphoneEncoder(FuzzyEncoder):
    """
    This is a fuzzy encoder that uses the
    `double metaphone <https://en.wikipedia.org/wiki/Metaphone>`_ algorithm.
    """
    def __init__(self):
        super().__init__(
            fn=self._dmetaphone
        )

    @classmethod
    def key(cls) -> str:
        """
        This is a key handle for this type of encoder.
        """
        return Encoders.DMETAPHONE.name

    @classmethod
    def _dmetaphone(cls, s: str):
        code = phonetics.dmetaphone(s)
        return code[0] if code else None


_ENCODER_TYPES = {
    **{
        k: v for k, v in {
            Encoders.SOUNDEX: SoundexEncoder,
            Encoders.NYSIIS: NysiisEncoder,
            Encoders.METAPHONE: MetaphoneEncoder,
            Encoders.DMETAPHONE: DMetaphoneEncoder
        }.items()
    },
    **{
        t.key(): t for t in [
            SoundexEncoder,
            NysiisEncoder,
            MetaphoneEncoder,
            DMetaphoneEncoder
        ]
    }
}  #: a mapping of encoder type keys to encoder types


def encoder(key: Encoders or str) -> Type:
    """
    Get a :py:class:`FuzzyEncoder` type that corresponds to a given enumeration
    value or name.

    :param key: the name
    :return: the corresponding type
    :raises KeyError: if no mapping exists for the key
    """
    return _ENCODER_TYPES[
        key.upper() if isinstance(key, str)
        else key
    ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Created by pat on 9/8/19
"""
.. currentmodule:: bnrml.soundalike
.. moduleauthor:: Pat Daburu <pat@daburu.net>

This package contains fuzzy string matching goodies.
"""
from .distance import (
    UnknownTextDistanceException,
    TextDistances,
    TextDistance,
    Hamming,
    Levenshtein,
    text_distance
)
from .encoding import (
    encoder,
    FuzzyEncoder,
    SoundexEncoder,
    NysiisEncoder,
    MetaphoneEncoder,
    DMetaphoneEncoder
)
from .normalizers import SoundalikeNormalsIndex



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. currentmodule:: bnrml
.. moduleauthor:: Pat Daburu <pat@daburu.net>

Be normal! (It's a string normalization framework.)
"""
from .normalizers import (
    load_normals,
    loadf_normals,
    SimpleNormalizer,
    Normal,
    Normalization,
    Normalizer
)
from .version import __version__, __release__
